#include "main.h"
#include "resource.h"

#pragma comment(lib, "comctl32.lib")


MAIN Main;

struct DirectoryNode
{
	DirectoryNode * Prev, * Next, * FirstChild, * Parent;

	char szShortName[MAX_PATH];
	signed __int64 nImmediateSize, nTotalSize;
};



int CALLBACK MainDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
int CALLBACK SubDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
int MapDir(const char * lpszBaseDir);
int MapDirR(DirectoryNode * ThisNode);
int FormatDirectoryName(char * lpszDirectoryName);
char * GetFullPath(DirectoryNode * Node);
int IncTotal(DirectoryNode * ThisNode, unsigned __int64 nAmt);
char * FormatSize(signed __int64 nSize);
int DeleteMap(DirectoryNode * BaseNode);
int PutMapIntoTree(HWND hTree, DirectoryNode * ThisNode, HTREEITEM ThisTreeItem);
int AssociateInRegistry();

DirectoryNode * HeadNode = NULL;
signed __int64 TotalDriveSize = 0, FreeDriveSize = 0;
bool HeadNodeIsDrive = true;
int bAutoSize = true;

int Startup()
{
	INITCOMMONCONTROLSEX iccs;
	iccs.dwSize = sizeof(INITCOMMONCONTROLSEX);
	iccs.dwICC = ICC_TREEVIEW_CLASSES;

	InitCommonControlsEx(&iccs);

	Main.hIcon = LoadIcon(Main.hInst, MAKEINTRESOURCE(IDI_ICON1));
	Main.hWnd = CreateDialog(Main.hInst, MAKEINTRESOURCE(IDD_MAIN_DLG), NULL, MainDlgProc);

	AssociateInRegistry();

	if (!Main.hWnd)
		return false;

	return true;
}

int Cleanup(int nExitValue)
{
	if (HeadNode)
		DeleteMap(HeadNode);

	DeleteObject(Main.hIcon);

	return nExitValue;
}

int StepTwo(char * lpszBaseDir)
{
	char szCaption[MAX_PATH + 32];
	sprintf(szCaption, "Mapping of %s", lpszBaseDir);

	ShowWindow(Main.hWnd, SW_HIDE);
	bAutoSize = IsDlgButtonChecked(Main.hWnd, IDC_AUTOSIZE);

	MapDir(lpszBaseDir);

	if (HeadNodeIsDrive)
	{
		strcat(szCaption, " (");
		strcat(szCaption, FormatSize(FreeDriveSize));
		strcat(szCaption, " / ");
		strcat(szCaption, FormatSize(TotalDriveSize));
		strcat(szCaption, " free)");
	}

	DestroyWindow(Main.hWnd);
	Main.hWnd = CreateDialog(Main.hInst, MAKEINTRESOURCE(IDD_DRIVE_SPACE), NULL, SubDlgProc);
	SetWindowText(Main.hWnd, szCaption);
	SetForegroundWindow(Main.hWnd);


	return true;
}

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR lpCmd, int nShow)
{
	Main.hInst = hInst;

	if (!Startup())
		return 0;

	if (strcmp(lpCmd, ""))
		StepTwo(lpCmd);
	else
		ShowWindow(Main.hWnd, nShow);


	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0))
	{
		if (!IsDialogMessage(Main.hWnd, &msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return Cleanup(msg.wParam);
}

int CALLBACK MainDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		CheckDlgButton(hWnd, IDC_AUTOSIZE, TRUE);
		SendMessage(hWnd, WM_SETICON, ICON_BIG, (LPARAM)Main.hIcon);
		SendMessage(hWnd, WM_SETICON, ICON_SMALL, (LPARAM)Main.hIcon);
		return TRUE;

	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_PICK_BASE_DIR:
			{
				BROWSEINFO bi;
				ZeroMemory(&bi, sizeof(BROWSEINFO));
				bi.hwndOwner = Main.hWnd;
				bi.lpszTitle = "Select a base folder from which to start mapping.";
				bi.ulFlags = BIF_RETURNONLYFSDIRS | BIF_DONTGOBELOWDOMAIN;

				ITEMIDLIST * pidl = SHBrowseForFolder(&bi);
				
				if (pidl)
				{
					char szBuffer[MAX_PATH];
					SHGetPathFromIDList(pidl, szBuffer);

					SetDlgItemText(Main.hWnd, IDC_BASE_DIR, szBuffer);
				}

			}
			break;

		case IDOK:
			{
				char szBaseDir[MAX_PATH];
				if (!GetDlgItemText(hWnd, IDC_BASE_DIR, szBaseDir, MAX_PATH))
				{
					MessageBox(hWnd, "You must have a base directory selected.", "User Error", MB_OK);
					SetFocus(GetDlgItem(hWnd, IDC_PICK_BASE_DIR));
				}
				else
					StepTwo(szBaseDir);
			}
			break;

		case IDCANCEL:
			PostQuitMessage(0);
			break;
		}
		break;

	
	}

	return FALSE;
}

int CALLBACK SubDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetWindowPos(hWnd, NULL, 0, 0, 300, 400, NULL);
		
		TreeView_SetBkColor(GetDlgItem(hWnd, IDC_TREE), RGB(0, 0, 0));
		TreeView_SetTextColor(GetDlgItem(hWnd, IDC_TREE), RGB(255, 255, 255));
		
		PutMapIntoTree(GetDlgItem(hWnd, IDC_TREE), HeadNode, NULL);

		SendMessage(hWnd, WM_SETICON, ICON_BIG, (LPARAM)Main.hIcon);
		SendMessage(hWnd, WM_SETICON, ICON_SMALL, (LPARAM)Main.hIcon);
		
		return TRUE;

	case WM_NOTIFY:
		{
			NMHDR * pnmh = (NMHDR*)lParam;
			if (pnmh->code == TVN_ITEMEXPANDED)
			{
				NMTREEVIEW * pnmtv = (NMTREEVIEW *)lParam;
				char szText[MAX_PATH];
				DirectoryNode * lpNode = (DirectoryNode *)pnmtv->itemNew.lParam;

				TVITEM tvi;

				tvi.mask = TVIF_TEXT;
				tvi.hItem = pnmtv->itemNew.hItem;
				tvi.pszText = szText;

				if (HeadNodeIsDrive)
				{
					if (pnmtv->action == TVE_EXPAND)
						sprintf(szText, "%s (%04.2f%%) [%s]", lpNode->szShortName, ((double)lpNode->nImmediateSize / (double)TotalDriveSize * 100), FormatSize(lpNode->nImmediateSize));
					else
						sprintf(szText, "%s (%04.2f%%) {%s}", lpNode->szShortName, ((double)lpNode->nTotalSize / (double)TotalDriveSize * 100), FormatSize(lpNode->nTotalSize));
				}
				else
				{
					if (pnmtv->action == TVE_EXPAND)
						sprintf(szText, "%s [%s]", lpNode->szShortName, FormatSize(lpNode->nImmediateSize));
					else
						sprintf(szText, "%s {%s}", lpNode->szShortName, FormatSize(lpNode->nTotalSize));
				}

				TreeView_SetItem(GetDlgItem(hWnd, IDC_TREE), &tvi);
			}
			else if (pnmh->code == NM_RCLICK)
			{
				TVHITTESTINFO tvhti;
				GetCursorPos(&tvhti.pt);
				ScreenToClient(GetDlgItem(Main.hWnd, IDC_TREE), &tvhti.pt);

				HTREEITEM HitItem = TreeView_HitTest(GetDlgItem(Main.hWnd, IDC_TREE), &tvhti);

				if (HitItem)
				{
					TVITEM tvi;
					tvi.hItem = tvhti.hItem;
					tvi.mask = TVIF_PARAM;
					TreeView_GetItem(GetDlgItem(Main.hWnd, IDC_TREE), &tvi);
					
					if (tvi.lParam)
					{
						DirectoryNode * Node = (DirectoryNode *)tvi.lParam;
						char szMessage[800];
						ZeroMemory(szMessage, 800);

						strcat(szMessage, "Full path: ");
						strcat(szMessage, GetFullPath(Node));
						strcat(szMessage, "\r\n");
						
						strcat(szMessage, "This node contains: ");
						strcat(szMessage, FormatSize(Node->nImmediateSize));
						strcat(szMessage, " bytes\r\n");
						
						if (Node->FirstChild)
						{
							strcat(szMessage, "This branch contains: ");
							strcat(szMessage, FormatSize(Node->nTotalSize));
							strcat(szMessage, " bytes.\r\n");
						}
						else
							strcat(szMessage, "This node has no children.\r\n");
						

						MessageBox(Main.hWnd, szMessage, "Directory Information", MB_OK);
					}
				}
				else
				{
					MessageBox(Main.hWnd, "Right click on an item to view more information about it.", "Information", MB_OK);
				}
			}
		}
		break;

	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
		case IDCANCEL:
			PostQuitMessage(0);
			break;
		}
		break;

	case WM_SIZE:
		SetWindowPos(GetDlgItem(hWnd, IDC_TREE), NULL, 0, 0, LOWORD(lParam), HIWORD(lParam), NULL);
		InvalidateRect(hWnd, NULL, FALSE);
		break;
	}

	return FALSE;
}

int MapDir(const char * lpszBaseDir)
{
	int nCt = 0;
	
	HeadNode = new DirectoryNode;
	ZeroMemory(HeadNode, sizeof(DirectoryNode));

	strcpy(HeadNode->szShortName, lpszBaseDir);
	FormatDirectoryName(HeadNode->szShortName);

	if ((isalpha(HeadNode->szShortName[0])) &&
		(HeadNode->szShortName[1] == ':') && 
		(HeadNode->szShortName[2] == '\0'))
	{
		HeadNodeIsDrive = true;

		HeadNode->szShortName[2] = '\\';
		HeadNode->szShortName[3] = '\0';


		GetDiskFreeSpaceEx(HeadNode->szShortName, NULL, (ULARGE_INTEGER *)&TotalDriveSize, (ULARGE_INTEGER *)&FreeDriveSize);

		HeadNode->szShortName[2] = '\0';
	}
	else
		HeadNodeIsDrive = false;


	nCt += MapDirR(HeadNode);

	return nCt;
}

int MapDirR(DirectoryNode * ThisNode)
{	// Used internally by MapDir...  Use MapDir(lpszBaseDir) instead
	int nCt = 0;

	DirectoryNode * CurNode;
	DirectoryNode * LastNode;

	WIN32_FIND_DATA wfd;
	char szFileMask[MAX_PATH];
	strcpy(szFileMask, GetFullPath(ThisNode));
	strcat(szFileMask, "*.*");

	HANDLE hFind = FindFirstFile(szFileMask, &wfd);

	if (hFind == INVALID_HANDLE_VALUE)
		return nCt;

	do
	{
		if ((!strcmp(wfd.cFileName, ".") || (!strcmp(wfd.cFileName, ".."))))
			continue;

		if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			CurNode = new DirectoryNode;
			ZeroMemory(CurNode, sizeof(DirectoryNode));
			CurNode->Parent = ThisNode;
			
			if (ThisNode->FirstChild == NULL)
				ThisNode->FirstChild = CurNode;
			else
			{
				LastNode->Next = CurNode;
				CurNode->Prev = LastNode;
			}

			sprintf(CurNode->szShortName, wfd.cFileName);

			nCt += MapDirR(CurNode);
			
			LastNode = CurNode;
		}
		else
		{
			unsigned __int64 FileSize = ((unsigned __int64)wfd.nFileSizeHigh * ((unsigned __int64)MAXDWORD + 1)) + (unsigned __int64)wfd.nFileSizeLow;

			IncTotal(ThisNode, FileSize);
			ThisNode->nImmediateSize += FileSize;
		}

		nCt++;
	}
	while (FindNextFile(hFind, &wfd));

	FindClose(hFind);

	return nCt;
}

int FormatDirectoryName(char * lpszDirectoryName)
{
	unsigned int n = strlen(lpszDirectoryName) - 1;

	if (lpszDirectoryName[n] != '\\')
		return false;
	
	lpszDirectoryName[n] = '\0';

	return true;
}

char * GetFullPath(DirectoryNode * Node)
{
	static char szPath[MAX_PATH];
	char szBuffer[MAX_PATH];
	ZeroMemory(szPath, MAX_PATH);

	while (Node)
	{
		strcpy(szBuffer, Node->szShortName);
		strcat(szBuffer, "\\");
		strcat(szBuffer, szPath);
		strcpy(szPath, szBuffer);
		Node = Node->Parent;
	}
	
	return szPath;
}

int IncTotal(DirectoryNode * ThisNode, unsigned __int64 nAmt)
{
	DirectoryNode * CurNode = ThisNode;
	int nCt = 0;

	while (CurNode)
	{
		CurNode->nTotalSize += nAmt;
		CurNode = CurNode->Parent;
		nCt++;
	}

	return nCt;
}

char * FormatSize(signed __int64 nSize)
{
	static char szSize[80];
	char szBuffer[80];
	char * lpszSize = szSize;
	char * lpszBuffer = szBuffer;

	if (bAutoSize)
	{
		double fSize = (double)nSize;
		char Demarkation[40];
		
		if (fSize >= (1024 * 1024 * 1024))
		{
			sprintf(Demarkation, " gigs");
			fSize /= (1024 * 1024 * 1024);
		}
		else if (fSize >= (1024 * 1024))
		{
			sprintf(Demarkation, " megs");
			fSize /= (1024 * 1024);
		}
		else if (fSize >= 1024)
		{
			sprintf(Demarkation, " k");
			fSize /= 1024;
		}
		else
			sprintf(Demarkation, " bytes");

		if (nSize >= 1024)
			sprintf(szSize, "%.1f", fSize);
		else
			sprintf(szSize, "%.0f", fSize);

		strcat(szSize, Demarkation);

	}
	else
	{
		sprintf(szBuffer, "%I64d", nSize);

		int i = strlen(szBuffer);
		int n = (i % 3) > 0 ? 3 - (i % 3) : 0;

		while (i)
		{
			if (n == 3)
			{
				*lpszSize = ',';
				lpszSize++;
				n = 0;
			}

			*lpszSize = *lpszBuffer;

			n++;
			i--;
			lpszSize++;
			lpszBuffer++;
		}

		*lpszSize = 0;

		strcat(szSize, " bytes");
	}

	return szSize;
}

int DeleteMap(DirectoryNode * BaseNode)
{
	int nCt = 0;

	DirectoryNode * CurNode = BaseNode;
	DirectoryNode * TmpNode;

	while (CurNode)
	{
		if (CurNode->FirstChild)
			nCt += DeleteMap(CurNode->FirstChild);

		TmpNode = CurNode->Next;
		delete CurNode;
		CurNode = TmpNode;
		nCt++;
	}

	return nCt;
}

int PutMapIntoTree(HWND hTree, DirectoryNode * ThisNode, HTREEITEM ThisTreeItem)
{
	int nCt = 0;
	DirectoryNode * CurNode = ThisNode;

	HTREEITEM CurTree;
	char szCaption[280];
	TVINSERTSTRUCT tvs;
	ZeroMemory(&tvs, sizeof tvs);

	while (CurNode)
	{
		nCt++;

		if (HeadNodeIsDrive)
			sprintf(szCaption, "%s (%04.2f%%) {%s}", CurNode->szShortName, ((double)CurNode->nTotalSize / (double)TotalDriveSize * 100), FormatSize(CurNode->nTotalSize));
		else
			sprintf(szCaption, "%s {%s}", CurNode->szShortName, FormatSize(CurNode->nTotalSize));
		
		tvs.hParent = ThisTreeItem;
		tvs.hInsertAfter = TVI_LAST;
		tvs.item.mask = TVIF_TEXT | TVIF_PARAM;
		tvs.item.pszText = szCaption;
		tvs.item.lParam = (long)CurNode;
		
		
		CurTree = TreeView_InsertItem(hTree, &tvs);

		if (CurNode->FirstChild)
			nCt += PutMapIntoTree(hTree, CurNode->FirstChild, CurTree);

		CurNode = CurNode->Next;
	}	


	return nCt;
}

int AssociateInRegistry()
{
	/*
	  Add key to HKEY_CLASSES_ROOT\Drive\shell  "Drive Space"
	  Add key to HKEY_CLASSES_ROOT\Drive\shell\Drive Space  "command"
	  Set default value to HKEY_CLASSES_ROOT\Drive\shell\Drive Space\command  "%PATH%\DriveSpace.exe %1"
	  Add key to HKEY_CLASSES_ROOT\Directory\shell  "Drive Space"
	  Add key to HKEY_CLASSES_ROOT\Directory\shell\Drive Space  "command"
	  Set default value to HKEY_CLASSES_ROOT\Directory\shell\Drive Space\command  "%PATH%\DriveSpace.exe %1"
	*/
	HKEY hReg[3];
	int ret = 0;
	char szAppCmd[400];

	GetModuleFileName(NULL, szAppCmd + 1, 400);
	szAppCmd[0] = '\"';
	strcat(szAppCmd, "\" %1");

	if (RegOpenKeyEx(HKEY_CLASSES_ROOT, "Drive\\shell", 0, KEY_ALL_ACCESS, &hReg[0]) == ERROR_SUCCESS)
	{
		ret++;
		if (RegCreateKeyEx(hReg[0], "Drive Space", NULL, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hReg[1], NULL) == ERROR_SUCCESS)
		{
			ret++;
			if (RegCreateKeyEx(hReg[1], "command", NULL, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hReg[2], NULL) == ERROR_SUCCESS)
			{
				ret++;
				RegSetValueEx(hReg[2], NULL, NULL, REG_SZ, (unsigned char *)szAppCmd, strlen(szAppCmd));				

				RegCloseKey(hReg[2]);
			}

			RegCloseKey(hReg[1]);
		}

		RegCloseKey(hReg[0]);
	}

	if (RegOpenKeyEx(HKEY_CLASSES_ROOT, "Directory\\shell", 0, KEY_ALL_ACCESS, &hReg[0]) == ERROR_SUCCESS)
	{
		ret++;
		if (RegCreateKeyEx(hReg[0], "Drive Space", NULL, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hReg[1], NULL) == ERROR_SUCCESS)
		{
			ret++;
			if (RegCreateKeyEx(hReg[1], "command", NULL, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hReg[2], NULL) == ERROR_SUCCESS)
			{
				ret++;
				RegSetValueEx(hReg[2], NULL, NULL, REG_SZ, (unsigned char *)szAppCmd, strlen(szAppCmd));				

				RegCloseKey(hReg[2]);
			}

			RegCloseKey(hReg[1]);
		}

		RegCloseKey(hReg[0]);
	}

	return ret;
}
